#ifndef BASICSOLVER_HPP_
#define BASICSOLVER_HPP_
#define basicsolver_hpp_version "$Revision: 1.35 $"

#include    <vector>

#include "ThreadGlobals.hpp"
#include 	"Assertion.hpp"
#include	"Types.hpp"
#include "Settings.hpp"
#include	"LazyFraction.hpp"
#include "Satistics.hpp"
#include "Heap.hpp"
#include "BitVector.hpp"
#include	"AssignInfos.hpp"
#include 	"ConflictsRestarts.hpp"
#include	"MemPoolA.hpp"
#include	"VecOfLists.hpp"
#include 	"Hash.hpp"
#include 	"IFStream.hpp"
#include 	"Logging.hpp"
#include	"WatchdCls.hpp"
#include	"HistoryBuffer.hpp"
#include <deque>

using std::vector;
using std::deque;

//#define RESOLVE_LEMMA_ZERO 1;
#define GC_LBD_MEDIAN 1


//typedef vector<ClsP> ::iterator cls_iter_t;
struct CoreCand{
	Literal 	lit;
	uint32_t	cls_id;
	CoreCand(const Literal& l, const uint32_t& id):lit(l),cls_id(id){}
	CoreCand(){}//dummy for vector.resize()
};





class BasicSolver{

protected:
	static const int32_t BCP_SUCCESS 	= -1; //SOMETHING < 0
	TGlobals&							glob;

    /** A heuristic measurement of the activity of a variable  */
	//Variable								next_ext_var;
	VarActComp							var_act_comp;
    var_hp_t   							var_act_hp;

    vector<Literal>						assign_stck;
public:
    VarVal								vals;
    Variable							sel_strt;
    bool								cls_mode;
protected:
    Assigns								ass_info;
    Watching							watching;

    uint32_t                        	queue_begin;	// Head of queue - as index into the assign_stack
    vector<uint32_t>                	dec_lev_end; 	// Indices for different decision levels in 'assign_stack'

    vector<Variable>					groups; //for group mode

    //uint32_t							real_units_end;	// End of real units as index into assign_stck


    uint32_t                        	simplify_assign_idx; // Number of top-level assignments at last execution of 'simplify()'
    uint64_t                         	simpl_next;	//Remaining number of propagations that must be made before next execution of 'simplify()'
   //bool								allow_gc;
    uint32_t							max_lbd_gc;
    bool								set_bcp_shrts;


    /** Data structure to set a flag for a variable and allow efficient clearing of all flags */
    BitVector<Variable>	            var_flags;

    LubyRestarts					luby_restarts;
    HistoryBuffer<uint8_t,uint64_t> lbd_history;
    uint64_t						lbd_total;
    uint64_t						lbd_confl;

    vector<Variable>				conflVars;

    //conflict related information
    vector<Literal>					lemma;//contains conflict lemma part with literals that are not frozen
    vector<Literal>					lemma_zro; //contains conflict lemma with frozen literals - may contain duplicate lits
    vector<Literal>					min_core;//contains unsat core -> entries are indices into assign_stack
    deque<CoreCand>					core_cands;
    BitVector<uint32_t>				confl_levs;
    uint32_t						confl_lbd;

    //reduction of lemma Sörensson Biere
    vector<Literal>                 reduce_lemma_stck;
    vector<Variable>				clear_later;//var flags to clear





//internal Methods

    //main
    TrBool			isSATInnr(const uint32_t max_strts);
    TrBool          search             	(uint32_t max_confl);
    Literal         nextDecision		(const bool rand );
    Literal         nextDecision		(){
					return nextDecision(ass_info.lessAgilThan(1000000) && glob.randy.nextInt(1000) < glob.sets.random_var_freq_permil); }


    //cleaners
    bool			shallSimplify()const{return simpl_next < glob.stats.cnt_props && assign_stck.size() != simplify_assign_idx;}
    void          	doSimplify          (const bool rearrange_clses, const bool bury_sat );
    void			cleanDB				( const uint32_t& max_lbd,  const uint32_t& lbd_rnd,
										const bool rearrange_clses,const bool bury_sat);
    uint32_t		clsRearrange(Cls& cls,WtchEle& wt, const Literal& ow, BitVector<Literal>* seen);
    void			setPureLits(BitVector<Literal>& seen);
    uint32_t		smpleChckUwSat(Cls& cls);
    bool			shallGC()const		{return (glob.stats.cnt_confl >= (glob.stats.cnt_gc+1)* glob.sets.max_learnts)
												|| glob.stats.litsCount() > glob.sets.max_lits ;}
    void 			learntsGC			();
    void			prepareNewRound(uint32_t next_simp)	{set_bcp_shrts = false; //queue_begin = 1;
    									simpl_next=glob.stats.cnt_props + next_simp;simplify_assign_idx=0;glob.reValidate(); }

    //propagation
    int32_t        	bcp           		( );
    int32_t        	bcpPropagateLit		(const Literal& false_lit);
    inline bool		bcpFindNewWatcher(const Literal& false_lit, const Literal& other, const WtchEle& curr_we);

    //set var values
	inline void     enqueue(const Literal& lit, const ClsP& from, const uint32_t& lbd);
	inline void     decide(const Literal& lit);
	inline void    	setUnit(const Literal& lit, const uint32_t& cls_id, const bool min_for_sure);

    //undo assignments in search
    inline void    	backjumpKeepLevel   (uint32_t level);    // Backtrack until a certain level.
    void    		backjumpUndoAssigns (uint32_t assign_idx); //undoes all assignments above and including the specified assign_idx


    //conflicts at higher level
    int32_t 		bcpFinishOnConflict(const Literal& false_lit, WIter<WtchEle>& iter, const ClsP& clp);
    void        	firstUIP            (int32_t nbr_pivot_lits);
    inline uint32_t firstUIPMrkLit		(const Literal& lit,const uint32_t& dec_level);
    void			resolveZeros(); //add literals of level 0 to lemma
    uint32_t        backtrackLevel      ();
    void            reduceSelfImpl  	();
    bool            litIsNecessary      (Literal lit);
    inline bool     reduceSelfImpCheckLiteral (const Literal& lit);
    inline void		clearLaterNow();

    //conflicts at level 0 and minimisation
    void			computeCore();
    void			clearAssignsHard(uint32_t from);//backjumps to level 0 and undoes all assignments down to assign_idx
    void			keepThisCandOnly(const Literal& cand);
    bool			propagateMinCore();//propagate while cand lits are not in assign_stck
    bool			reviveClauses();


    //heuristic
    inline void     varDecayActivity();
    inline void		varBump				(const Variable& var);

    //aux
	uint32_t  decisionLevel()  const 		{return dec_lev_end.size();}
	bool setLevel(const uint32_t& l)		{asrt(l<=confl_levs.size()); return confl_levs.setFlagNew(l);}
	void zeroLevel(const uint32_t& l)		{asrt(l<=confl_levs.size()); confl_levs.carelessClear(l);}
	bool isSetLevel(const uint32_t& l)const	{asrt(l<=confl_levs.size()); return confl_levs.isSet(l);}
	void setVarOf(const Literal& lit)		{var_flags.setFlag(lit.toVar());}
	uint32_t		getLBD(const Cls& cls, const WtchEle& we);
	//uint32_t		getLBD(const WtchEle& we){return getLBD(glob.getCls(we.getClsP()),we);}

    //initialisation
    ClsP			clsCreate(vector<Literal>& lits, const uint32_t ufix,const bool learnt ,uint32_t lbd, bool set_shrt);
    inline bool 	haveToAllocVars(uint32_t max_idx){return max_idx >= var_flags.size();}
    inline void   	reserveForVariables(uint32_t max_idx,const bool in_heap);

    inline void 	reserveForClauses(uint32_t need_sz);
	bool    		addPrsdClause(vector<Literal>& lits, const uint32_t cls_id, Variable& sel_var);
	uint32_t		validateCls(vector<Literal>& lits);
	bool 			parseRegular(IFStream& in, const uint32_t vars, const uint32_t cls);
	bool 			parseGroups(IFStream& in,const uint32_t vars, const uint32_t expect);

	//this is for the world
public:
                    BasicSolver(TGlobals& g);
                    ~BasicSolver();

	TrBool 			isSAT();
	bool 			parseDIMACS(IFStream& in);
	void			printModel(std::ostream& out);
	void			minimiseCore();
	void			printCore(vector<uint32_t>* cls_ids);

private:
	//bool			asrtAllFalse(Literal* from, Literal* end);
	bool			asrtNoBadLevels();
	bool			asrtClsUnit(const Literal& u,const ClsP& clp);
	bool			asrtClsDontContain(const Literal& wtch, const ClsP& clp, const Literal& cand);
	bool			asrtNotContained(const Literal& cand);
	bool			asrtAllFalse(const Literal& wtch, const ClsP& clp, const uint32_t max_lev);
	bool			asrtOnlyInitialUnitsEnqueued();
	bool			asrtUnitsEnquedOrInCoreCands();
	bool			asrtVarEnqued(const Variable& var)const;
	bool			asrtVarInCoreCands(const Variable& var)const;
	bool			asrtAllClsesSAT()const;
	bool			asrtPrintAllClses() const;

public:
	bool 			asrtVerifyModel(const char* file);

};



//Inline Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void	BasicSolver::varBump(const Variable& var){
	var_act_hp          .activityBump(var);
}

void	BasicSolver::clearLaterNow(){
   	for(uint32_t i=clear_later.size();i--;) {
   		var_flags.carelessClear(clear_later[i]);}
   	clear_later.clear();
}



void    BasicSolver::setUnit(const Literal& lit, const uint32_t& cls_id, const bool min_for_sure){
	asrt(vals.litUkwn(lit)); asrt(decisionLevel()==0); //asrt(real_units_end == assign_stck.size());
	LOG2("Set unit literal " << lit.toDIMACS() << " with id " << cls_id << " for sure is " << min_for_sure);

	ass_info	.setValIniUnit(lit,cls_id,min_for_sure); //asrt(ass_info.reasonLBD(lit.toVar()) == Reason::LBD_MAX);
    assign_stck	.push_back(lit);
}

void     BasicSolver::enqueue(const Literal& lit, const ClsP& from, const uint32_t& lbd){
	asrt(vals.litUkwn(lit)); asrt(asrtClsUnit(lit,from));
	LOG2("enqueue " << lit.toDIMACS() << " at level " << decisionLevel());

	ass_info.setValueReason(lit,decisionLevel(),from,lbd); asrt(ass_info.reasonLBD(lit.toVar()) == lbd);
	assign_stck.push_back(lit);
}

void    BasicSolver::decide(const Literal& lit){ sanity(confl_levs.allCleared());  asrt(vals.litUkwn(lit));

	dec_lev_end	.push_back(assign_stck.size());
	confl_levs	.ensureSize(dec_lev_end.size()+1,false);

    ass_info	.setValueDecision(lit,decisionLevel()); asrt(ass_info.reasonLBD(lit.toVar()) == Reason::LBD_MAX);
    assign_stck	.push_back(lit);
}




void     	BasicSolver::reserveForVariables(uint32_t max_idx,const bool in_heap){

	if(++max_idx > var_flags.size()){
	    LOG2("reserve for variables " << max_idx);

	    vals			.ensureVarVals(max_idx);
	    ass_info		.ensureVars(max_idx);
	    var_flags		.ensureSize(max_idx,false);
	    glob			.ensureVarVals(max_idx);
	    if(in_heap){
	    	var_act_hp	.insertElementsTo(max_idx,glob.sets.var_min_inc);
	    }else{
	    	var_act_hp	.ensureElementValsTo(max_idx,glob.sets.var_min_inc);
	    }

	    max_idx <<= 1;
	    watching		.enlarge(max_idx);
	 }
}

void BasicSolver::reserveForClauses(uint32_t need_sz){
	static const uint32_t avg_cls_len = 2;

	glob.c_info	.reserveFor(need_sz * avg_cls_len);
	watching	.reserveForTotal(need_sz * 2);
}

void    	BasicSolver::backjumpKeepLevel(uint32_t level){

    if(decisionLevel() <= level) return;
    backjumpUndoAssigns(dec_lev_end[level]);

    dec_lev_end    	.resize(level);
}





/*
void BasicSolver::compCoreCheckLit(const Literal& lit){
   const Variable      var(lit.toVar());    asrt(vals.litFalse(lit)); asrt(decisionLevel() == 0); asrt(ass_info.varAssLev(var)==0);

   if(var_flags.isSet(var)) return;

   var_flags.setFlag(var);

   if(ass_info.varAssignReason(var).isUnit()){
	   lemma_zero.push_back(lit);
   }
}*/



uint32_t BasicSolver::firstUIPMrkLit(const Literal& lit,const uint32_t& dec_level){
   const Variable      var(lit.toVar());    asrt(vals.litFalse(lit)); asrt(dec_level > 0);
   LOG2("firstUIPMrk check lit " << lit.toDIMACS());
   if(var_flags.isSet(var)) return 0 ;

   uint32_t level 		= ass_info.varAssLev(var);
   LOG2("firstUIPMrk lit " << lit.toDIMACS() << " of level " << level << " curr is " << dec_level);

   if(level == 0){		LOG2("Put lit in lemma_zro " << lit.toDIMACS());
	   lemma_zro		.push_back(lit); //allow duplicates
	   return 0;
   }

   var_flags			.setFlag(var);
   varBump(var);		LOG2("Bump var " << var.toDIMACS());

   if(level ==  dec_level){

	   //glucose
	   if(glob.sets.bump_glucose){conflVars.push_back(var);}
	   return 1;
   }

   LOG2("Put lit in lemma " << lit.toDIMACS());
   lemma			.push_back(lit);
   confl_lbd		+= setLevel(level);
   return 0;
}


bool BasicSolver::reduceSelfImpCheckLiteral(const Literal& lit){

    Variable var = lit.toVar();   asrt(var.idx() < var_flags.size());

    if(var_flags.isSet(var)) return false;

    const uint32_t lev	= ass_info.varAssLev(var);

    if(lev == 0){
    	lemma_zro			.push_back(lit); //allow duplicates
    	return false;
    }


    if(ass_info.varIsDecision(var) || !isSetLevel(lev)){
    	return true;}

    reduce_lemma_stck	.push_back(lit);
    var_flags			.setFlag(var);
    return false;
}


bool	BasicSolver::bcpFindNewWatcher(const Literal& false_lit, const Literal& other, const WtchEle& curr_we){

	Cls& cl 		= glob.getCls(curr_we.getClsP()); asrt(false_lit != other); asrt(other == cl.otherWatcher(false_lit));
	ClsLit* clit 	= cl.uwBeg();
	do{
		if(!vals.litFalse(Literal(clit->lit))){
			cl			.changeWatchers(other,clit,false_lit); LOG2("Found new watcher "<<cl.otherWatcher(other).toDIMACS());
			watching	.changeWatching(curr_we,other,cl.otherWatcher(other));
			return true;
		}
	}while((clit++)->propNext());

	return false;
}





const char*  basicsolver_cpp_version();
const char*  main_cpp_version();
#endif /*BASICSOLVER_HPP_*/
