
#ifndef SATISTICS_HPP_
#define SATISTICS_HPP_
#define satistics_hpp_version      "$Revision: 1.13 $"


#include <stdio.h>
#include <stdarg.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string>
#include "Types.hpp"

using std::string; using std::cout; using std::endl;

//#define EXP_TIME 1
#define COLLECT_STATS 1

#ifdef COLLECT_STATS
#define opt_istat(v,i) glob.stats.nbr_##v += i;
#define opt_sstat(v,i) glob.stats.nbr_##v = i;
#else
#define opt_istat(v,i) asrt(true);
#define opt_sstat(v,i) asrt(true);
#endif

#define cnt_stat(v,i) glob.stats.cnt_##v += i;

class SimplyFire;

class Satistics{
	template<class CP>
	inline void printStats(const std::string& result, CP& p);

	inline double			cpuUsrUTime();
	inline double			cpuSysTime();
public:
	inline uint64_t 		usedMemoryBytes();
	inline double 			cpuUsrTime();
private:

	const int 				pid;
	const uint64_t			page_sz;
	const string			prgrss_strt;
	const string			prgrss_dec;
	const string			prgrss_props;
	const string			prgrss_confl;
	const string			prgrss_lits;
	const string			prgrss_gc;
	const uint32_t			prgrss_head_sz;


public:
//statistics that are used for heuristics
	uint64_t				cnt_lts;
	uint64_t				cnt_props;
	uint32_t				cnt_strts;
	uint32_t				cnt_act_decs;
	uint32_t				cnt_rnd_decs;
	uint32_t				cnt_confl;
	uint32_t				cnt_gc;
	uint32_t				cnt_bin_cls;
	//uint32_t				cnt_trn_cls;
	uint32_t				cnt_lcls;
	uint32_t				cnt_core_lts;

	uint32_t				nbr_orig_vars;
	uint32_t				nbr_orig_cls;
	uint32_t				nbr_orig_bins;
	//uint32_t				nbr_orig_terns;
	uint32_t				nbr_orig_lits;


#ifdef COLLECT_STATS
//statistics just for fun

	uint64_t				nbr_sum_confl_depth;
	uint64_t				nbr_sum_bck_jmp;
	uint32_t				nbr_simpl_sml;
	uint32_t 				nbr_rstrts_skip;


	//uint32_t				nbr_cls_one_diff_fail;
	//uint32_t				nbr_cls_one_diff_work;
	//uint32_t				nbr_ext_var_exists;

	uint32_t				nbr_subchk;
	uint32_t				nbr_bwsub_litselim;
	uint32_t				nbr_bwsub_self;
	uint32_t				nbr_sub_work;

	//uint32_t				nbr_esub_self;
	//uint32_t				nbr_esub_selfbw;
	//uint32_t				nbr_esub_other;
	//uint32_t				nbr_esub_otherbw;

	uint32_t				nbr_assym_clschk;
	uint32_t				nbr_assym_litsrem;
	uint32_t				nbr_assym_dom;
	uint32_t				nbr_assym_shrnk_easy;
	uint32_t				nbr_assym_shrnk_ana;
	uint32_t				nbr_assym_subsum_easy;

	uint32_t				nbr_pure_lits;

	//uint32_t				nbr_ext_vars_ass;
	//uint32_t				nbr_ext_vars_cycle;
#endif

#ifdef EXP_TIME
//time
	double					tim_parse;
	double					tim_prop;
	double					tim_conflana;

	//double				tim_eq_but_one;
	double					tim_simpl;
	double					tim_gc;
	double					tim_subchk;
	double					tim_assymchk;
	double					tim_assymbcpbin;
	double					tim_assymbcp;
	double					tim_assymflscls;
	double					tim_assymtouchdcls;

#endif

	Satistics():
		pid(getpid()),
		page_sz(getpagesize()),
		prgrss_strt("     Starts:  [s]"),
		prgrss_dec("  Decisions  "),
		prgrss_props(" Propagations "),
		prgrss_confl("  Conflicts  "),
		prgrss_lits("     Literals:    [MB]"),
		prgrss_gc(" GC Calls "),
		prgrss_head_sz(prgrss_strt.size()+ prgrss_dec.size()+ prgrss_props.size()+ prgrss_confl.size()+prgrss_lits.size()+ prgrss_gc.size()+9),


		cnt_lts(0)
		,cnt_props(0)
		,cnt_strts(0)
		,cnt_act_decs(0)
		,cnt_rnd_decs(0)
		,cnt_confl(0)
		,cnt_gc(0)
		,cnt_bin_cls(0)
		//,cnt_trn_cls(0)
		,cnt_lcls(0)
		,cnt_core_lts(0)


		,nbr_orig_vars(0)
		,nbr_orig_cls(0)
		,nbr_orig_bins(0)
		//,nbr_orig_terns(0)
		,nbr_orig_lits(0)

#ifdef COLLECT_STATS
		,nbr_sum_confl_depth(0)
		,nbr_sum_bck_jmp(0)
		,nbr_simpl_sml(0)
		,nbr_rstrts_skip(0)


		//nbr_cls_one_diff_fail(0)
		//nbr_cls_one_diff_work(0)
		//nbr_ext_var_exists(0)

		,nbr_subchk(0)
		,nbr_bwsub_litselim(0)
		,nbr_bwsub_self(0)
		,nbr_sub_work(0)


		,nbr_assym_clschk(0)
		,nbr_assym_litsrem(0)
		,nbr_assym_dom(0)
		,nbr_assym_shrnk_easy(0)
		,nbr_assym_shrnk_ana(0)
		,nbr_assym_subsum_easy(0)

		,nbr_pure_lits(0)

		//nbr_ext_vars_ass(0)
		//nbr_ext_vars_cycle(0)
#endif

#ifdef EXP_TIME
		,tim_parse(0)
		,tim_prop(0)
		,tim_conflana(0)

		//tim_eq_but_one(0)
		,tim_simpl(0)
		,tim_gc(0)
		,tim_subchk(0)
		,tim_assymchk(0)
		,tim_assymbcpbin(0)
		,tim_assymbcp(0)
		,tim_assymflscls(0)
		,tim_assymtouchdcls(0)
#endif
	{}

	inline void printProgress();
	inline void printProgressHead();
	inline void printCSV(std::string file_name, TrBool result);
	inline void printStatsConsole(TrBool result);
	uint64_t	litsCount() const {return cnt_lts;}
	void keepOrig(const uint32_t vrs){
		nbr_orig_vars 	= vrs;
		nbr_orig_cls	= cnt_lcls;
		nbr_orig_bins	= cnt_bin_cls;
		//nbr_orig_terns	= cnt_trn_cls;
		nbr_orig_lits	= cnt_lts;
	}


	//Conflicts& getConflictCount()				{return conflict_info;}

};


double Satistics::cpuSysTime(){
	struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_stime.tv_sec + (double)ru.ru_stime.tv_usec / 1000000.0;
}


double Satistics::cpuUsrUTime(){
	struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec*1000000.0 + (double)ru.ru_utime.tv_usec ;
}

double Satistics::cpuUsrTime(){
	return cpuUsrUTime() / 1000000.0;
}



uint64_t Satistics::usedMemoryBytes(){
    char    path[256];
    sprintf(path, "/proc/%d/statm", pid);
    FILE*   in = fopen(path, "rb");
    if (in == NULL) return 0;
    uint64_t     value;
# if __WORDSIZE == 64
    if(!fscanf(in, "%ld", &value)){value = 0;}
#else
    if(!fscanf(in, "%lld", &value)){value = 0;}
#endif
    fclose(in);
    return value * page_sz;
}




#ifdef EXP_TIME
#define T_ON(w)  glob.stats.tim_##w -= glob.stats.cpuUsrTime()
#define T_OFF(w) glob.stats.tim_##w += glob.stats.cpuUsrTime()
#else
#define T_ON(what)  asrt(true)
#define T_OFF(what) asrt(true)
#endif



inline std::string toString(double d){
	static char buff[30];
	sprintf(buff," %*.3f ",13,d);
	return std::string(buff);
}

inline std::string toString(uint64_t l){
	static char buff[30];

# if __WORDSIZE == 64
	sprintf(buff," %13lu ",l);
#else
    sprintf(buff," %13llu ",l);
#endif

	return std::string(buff);
}

inline std::string toString(uint32_t i){
	static char buff[30];
	sprintf(buff," %13u ",i);
	return std::string(buff);
}


inline std::string toString(int i){
	static char buff[30];
	sprintf(buff," %13i ",i);
	return std::string(buff);
}

inline std::string toString(const std::string& s){
	if(s.size()>= 15) return s;
	std::string ret = "                " + s;
	return ret.substr(ret.size()-15,15);
}


void Satistics::printProgressHead(){

	//string name("")

	cout 	<< "c " << std::setw(prgrss_head_sz-2) << std::setfill('c') << "c" << endl
			<< "c " << prgrss_strt
			<< "|"	<< prgrss_dec
			<< "|"	<< prgrss_props
			<< "|"	<< prgrss_confl
			<< "|"	<< prgrss_lits
			<< "|"	<< prgrss_gc
			<< " c"	<< endl;
}


void Satistics::printProgress(){
	const uint32_t ut 	= (uint32_t)(10.0 * cpuUsrTime());
	const uint32_t mem	= (uint32_t)(10.0 * (usedMemoryBytes()/MEGA)) ;

	cout 	<< "c " << std::setfill(' ')
			<< std::setw(prgrss_strt.size()-8) << cnt_strts << ":" << std::setw(5) <<ut/10 <<"."<<ut%10 <<"|"
			<< std::setw(prgrss_dec.size()) << cnt_act_decs + cnt_rnd_decs << "|"
			<< std::setw(prgrss_props.size()) << cnt_props << "|"
			<< std::setw(prgrss_confl.size()) << cnt_confl << "|"
			<< std::setw(prgrss_lits.size()-9) << litsCount() << ":" << std::setw(6)  <<mem/10<<"."<<mem%10 <<"|"
			<< std::setw(prgrss_gc.size()) << cnt_gc << " c" << endl;
}


struct CSVPrinter{
	std::ofstream& out;
	std::string values;
	CSVPrinter(std::ofstream& o):out(o){}
	void head(const std::string& s)	{out << std::setw(15) << s << " & ";}
	void val(const std::string& s)	{values += s + " & ";}
	void skipLines(uint32_t l)		{};
	void finish(){
		time_t rawtime; time(&rawtime); struct tm* timeinfo = localtime(&rawtime);
		char* date_time = asctime(timeinfo); *(date_time + strlen(date_time) - 1) = ' ';
		head("Date"); 	val(std::string(" \"" + std::string(date_time) + "\" "));
		head("END"); 	val("END");
		out <<  endl << values << endl;}
};

struct ConsolePrinter{
	const uint32_t width;
	ConsolePrinter(uint32_t w):width(w-43){}
	void head(const std::string& s)	{cout << "c " << std::setw(25) << s << ":";}
	void val(const std::string& s) 	{cout << std::setw(15) << s << std::setw(width) << "c" << endl;}
	void finish()					{};
	void skipLines(uint32_t l)		{};
};

void 	Satistics::printCSV(std::string file_name, TrBool result){
	asrt(!file_name.empty());
	string res			= (result.isKnown() ? (result.isTrue()?"SAT":"UNSAT") : "UNKNOWN");
	res += "               ";
	std::ofstream out(file_name.c_str());
	CSVPrinter printer(out);
	printStats(res,printer);
}


#define VALUE(x) p.val(toString(x))

void Satistics::printStatsConsole(TrBool result){
	string res				= result.isKnown() ? (result.isTrue()?"SAT ":"UNSAT ") : "UNKNOWN ";
	cout << "c " << std::setw(prgrss_head_sz-2) << std::setfill('c') << "c" << endl;
	cout << std::setfill(' ');
	ConsolePrinter p(prgrss_head_sz);
	printStats(res,p);

	cout << "c " << std::setw(prgrss_head_sz-2) << std::setfill('c') << "c" << endl;
	cout << std::setfill(' ');
}

#include "ThreadGlobals.hpp"
#define a_per_b(x,t) 	((t)==0?0.0:((double)(x)/(double)(t)))
#define perc(x)	a_per_b	((x*100.0),timestmp)
template<class CP>
void Satistics::printStats(const std::string& result, CP& p){

	const double  timestmp 			= cpuUsrTime();

    std::string values  = "";

    p.head("User Time [s]");		VALUE(timestmp);
    p.head("System Time [s]");		VALUE(cpuSysTime());
    p.head("Memory [MB]");			VALUE((double)usedMemoryBytes() / (double)MEGA);
    p.head("Result");				VALUE(result);

    p.head("Orig Vars");			VALUE(nbr_orig_vars);
    p.head("Orig long Clses");		VALUE(nbr_orig_cls);
    p.head("Orig bin Clses");		VALUE(nbr_orig_bins);
    p.head("Orig Lits");			VALUE(nbr_orig_lits);
    p.head("Orig Lits per cls");	VALUE((double)(nbr_orig_lits)/(double)(nbr_orig_bins+nbr_orig_cls));

    p.head("Bin Clses");			VALUE(cnt_bin_cls);
    p.head("Long Clses");			VALUE(cnt_lcls);
    p.head("Literals");				VALUE(litsCount());

    p.head("Starts");				VALUE(cnt_strts);
    p.head("Propagations");			VALUE(cnt_props);
    p.head("Act Decisions");		VALUE(cnt_act_decs);
    p.head("Rnd Decisions");		VALUE(cnt_rnd_decs);
    p.head("Props per Decision");	VALUE((double)cnt_props/(double)(cnt_rnd_decs+cnt_act_decs));
    p.head("Conflicts");			VALUE(cnt_confl);
    p.head("Props per Conflict");	VALUE((double)cnt_props/(double)cnt_confl);
    p.head("Decs per Conflict");	VALUE((double)(cnt_act_decs+cnt_rnd_decs)/(double)cnt_confl);
    p.head("Garbage Collect");		VALUE(cnt_gc);
    p.head("Final Bin Clses");		VALUE(cnt_bin_cls);
    p.head("Final Lits per cls");	VALUE((double)(cnt_lts)/(double)(cnt_bin_cls+cnt_lcls));

    p.head("Size of UNSAT Core");	VALUE(cnt_core_lts);

#ifdef COLLECT_STATS




    p.head("Simplify");				VALUE(nbr_simpl_sml);
    p.head("Sum Backjump");			VALUE(nbr_sum_bck_jmp);
    p.head("Avg Backjump");			VALUE((double)nbr_sum_bck_jmp / (double)cnt_confl);
    p.head("Avg Confl Dep");		VALUE((double)nbr_sum_confl_depth / (double)cnt_confl);

    p.head("Pure Literals");		VALUE(nbr_pure_lits);

    p.head("Subsum Checks");		VALUE(nbr_subchk);
    p.head("Subsum Works");			VALUE(nbr_sub_work);
    p.head("BWSub Self subsum");	VALUE(nbr_bwsub_self);
    p.head("BWSub Rem. Lits");		VALUE(nbr_bwsub_litselim);


    p.head("Assym cls check");		VALUE(nbr_assym_clschk);
    p.head("Assym total shrink");	VALUE(nbr_assym_litsrem);
    p.head("Assym bin add");		VALUE(nbr_assym_dom);
    p.head("Assym shrink easy");	VALUE(nbr_assym_shrnk_easy);
    p.head("Assym shrink mdum");	VALUE(nbr_assym_shrnk_ana);
    p.head("Assym subsum easy");	VALUE(nbr_assym_subsum_easy);
#else

    //p.head("Define COLLECT_STATS at compilation to get more statistics");	VALUE(0);
#endif


    //p.head("Vars");					VALUE(result);



#ifdef EXP_TIME
    p.head("Expensive Profiling [s]");	VALUE("---");
    p.head("Parse Time");			VALUE(tim_parse);
    p.head("BCP");					VALUE(tim_prop);
    p.head("Conflict Ana");			VALUE(tim_conflana);
    //p.head("One Diff time [s]");	VALUE(tim_eq_but_one);
    //p.head("One Diff %");			VALUE(perc(tim_eq_but_one));

    p.head("Simplification");		VALUE(tim_simpl);
    p.head("Garbage Collection");	VALUE(tim_gc);
    p.head("Subsum Checks");		VALUE(tim_subchk);
    p.head("Assym Totl");			VALUE(tim_assymchk);
    p.head("Assym BCP Bin");		VALUE(tim_assymbcpbin);
    p.head("Assym BCP");			VALUE(tim_assymbcp);
    p.head("Assym False Cls ");		VALUE(tim_assymflscls);
    p.head("Assym Touched Cls");	VALUE(tim_assymtouchdcls);
#endif
    //p.head("Max Active Vars");		VALUE(var_act_hp.active_count);
    //p.head("Act Vars Percent");		VALUE(((double)var_act_hp.active_count*100.0)/((double)nVariables()));

    p.finish();
}





#endif /* SATISTICS_HPP_ */
