#ifndef SETTINGS_HPP_
#define SETTINGS_HPP_
#define settings_hpp_version "$Revision: 1.5 $"

#include "Random.hpp"
#include "ThreadGlobals.hpp"
#include "Types.hpp"
#include "OFStream.hpp"
#include <stdio.h>
#include "Assertion.hpp"
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <string>
#include <utility>
#include <string>
#include <iostream>
#include <iomanip>
#include "EInfo.hpp"
using std::string; using std::cout; using std::endl;



/**
 * Printing Functionality
 *
 */
//#define print_out(format, args...) ( fflush(stdout), fprintf(stderr, format, ## args), fflush(stderr) )

typedef unsigned int uint128_t __attribute__((mode(TI)));

template<class T>
struct Option{
    typedef std::pair<const char*,T>           val_type;
private:
	T* 									       variable;
	const char*                                command_arg;
	int                                        command_len;
	const char*                                description;
	std::vector<val_type >                     possible_vals;

public:

	Option(T* var,T default_val,const char* arg,const char* descr)
		:variable(var),command_arg(arg),command_len(strlen(arg)),description(descr){
	    *variable      = default_val;};

	void addPossibleValue(val_type pval){  possible_vals.push_back(pval);}

	const char* canHandle(const char* arg){
	      return (strncmp(arg, command_arg, command_len) == 0) ? (arg + command_len) : NULL;}

	void assign(const T val){*variable = val;}

	bool findValue(const char* arg){
	    if(possible_vals.empty()) return false;
	    for(int i = possible_vals.size();i--;){
	        if(strcmp(possible_vals[i].first,arg) == 0){
	            assign(possible_vals[i].second);
	            return true;
	        }
	    }
	    cout << "Unknown value " << arg << " of option " << command_arg << endl;
	    exit(EXIT_FAILURE);
	    return false;
	}

	void printValue(){
	    for(int i = possible_vals.size();i--;){
	        if(*variable == possible_vals[i].second){
	            cout << std::setw(20) << command_arg <<"\t "<< possible_vals[i].first << endl;
	            return;
	        }
	    }
        cout << std::setw(20) << command_arg << "\t "<< *variable << endl;
	}


	void printDescription(){
	    cout << std::setw(20)  << command_arg <<"\t "<< description;
	    int vals = possible_vals.size();
	    if(vals){
	        cout << " - possible values are: ";
	        for(int i = 0; i != vals;++i){ cout << " | " << possible_vals[i].first;}
	    }
	    cout << endl;
	}
};




struct Settings{
	bool				print_console;
	bool				print_values;
	bool				minimise;

	uint             	rand_seed;
	uint32_t			pol_mode;

	uint32_t			restart_type;
	static const uint32_t RSTRT_GLUE = 0;
	static const uint32_t RSTRT_LUBY = 1;

	bool				bump_agil;
	bool 				bump_glucose;

	uint32_t			lbd_rest_buffr;
    uint32_t         	timeout;

    uint32_t			max_learnts;
    uint32_t			max_learnts_inc;
    uint32_t			max_lits;

    string            cnf_file_name;
    string            csv_file_name;
    string            out_file_name;

    string            solver_name;

    const var_act_t   var_decay;       // Inverse of the variable activity decay factor.
    const var_act_t   var_min_inc;
    const var_act_t   var_max_act;

    uint32_t		  	random_var_freq_permil;

    std::vector<Option<uint> >   int_options;
    std::vector<Option<bool> >  bool_options;
    std::vector<Option<string> > string_options;

    Settings();
    ~Settings();

    bool             read(int argc, char** argv);
    bool             read(const char* arg);
    void             printUsage();
    void			 printCurrSettings();
    void			 printSizes();
    void 			 printCVSVersionNumbers();
};




//inline int min_int(const int x, const int y)            { return (y + (-(x < y) & (x - y)));}

//inline int max_int(const int x, const int y)            { return (y + (-(x > y) & (x - y)));}

//inline void set_max_int(uint32_t& to_be_max, const uint32_t cand) { to_be_max += (-(to_be_max < cand) & (cand - to_be_max));}

//inline void set_min_int(uint32_t& to_be_min, const uint32_t cand) { if(cand < to_be_min){to_be_min = cand;}}
inline void set_max_int(uint32_t& to_be_max, const uint32_t cand) { if(cand > to_be_max){to_be_max = cand;}}
//inline void swap(int& x, int& y)                        { x ^= y; y ^= x; x ^= y;}




template<class E, class Pred>
inline bool find(const std::vector<E>& vec,Pred p){

	int32_t i = vec.size();
    while(i--){if(p(vec[i])){ return true;}}

    return false;
}


template<class E>
inline uint32_t find(const std::vector<E>& vec,const E& e){

	int32_t i = vec.size();
    while(i--){if(vec[i]==e){ return i;}}

    return max32;
}


template<class E>
inline uint32_t find(E* vec,uint32_t sz,const E& e){

    while(sz--){if(vec[sz]==e){ return sz;}}
    return max32;
}


template<class E>
inline uint32_t common(const std::vector<E>& a, const std::vector<E>& b){
	uint32_t i = 0; const uint32_t sz = a.size() > b.size() ? b.size() : a.size();
    for(;i!=sz;++i){if(a[i] != b[i]){ break;}}
    return i;
}


inline bool powOf2(const uint64_t x){
	//cout << " check " << x << endl;
	return (x & -x) == x;
}



template<class E, class Pred>
inline void stable_remove_one(std::vector<E>& vec,Pred p){
    asrt(!vec.empty());

    int32_t i = vec.size();
    while(i--){if(p(vec[i])){ ++i; break;}}
    asrt( i > 0 );

    for(;i < (int32_t)vec.size();++i){vec[i-1]=vec[i];}
    vec.pop_back();
}


template<class E, class P>
inline void unstable_remove_one(std::vector<E>& vec, P p){
    asrt(!vec.empty());

    for(int32_t i(vec.size());i--;){
        if(p(vec[i])){
            vec[i] = vec.back(); vec.pop_back();
            return;
    }}
    asrt(false);
}

template<class Vec>
inline uint32_t sum_element_sizes(const Vec& vec){
    uint32_t sum = 0;

    for(uint32_t i = 0;i != vec.size(); ++i){
    	sum += vec[i].size();
    }

    return sum;
}

inline char* bname(char* str, char sep = '/'){
    for(char* iter = str; *iter != '\0' ; ){
        if(*iter++ == sep) str = iter;
    }
    return str;
}



template<class T>
inline uint32_t bin_search(const T eles[],uint32_t st, uint32_t sz,const T& e){
	sanity(is_sorted(eles,st,sz));
	while(st<sz){
		const uint32_t p 	= (st+sz)>>1; asrt(p<sz);asrt(p>=st);

		if(eles[p]<e)		{st	=p+1;}
		else if(e<eles[p])	{sz	=p;}
		else {				return p;}
	}
	return max32;
}



const char*  settings_cpp_version();
const char*  main_cpp_version();
#endif /*SETTINGS_HPP_*/
